from typing import Callable, Iterable

import click
from click import Group
from click.decorators import F
from click_repl import ClickCompleter, text_type
from prompt_toolkit.completion import CompleteEvent, Completion
from prompt_toolkit.document import Document

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice
from i8_terminal.types.chart_param_type import ChartParamType
from i8_terminal.types.command_parser import CommandParser
from i8_terminal.types.fin_identifier_param_type import FinancialsIdentifierParamType
from i8_terminal.types.fin_period_param_type import FinancialsPeriodParamType
from i8_terminal.types.fin_statement_param_type import FinancialStatementParamType
from i8_terminal.types.indicator_param_type import IndicatorParamType
from i8_terminal.types.metric_param_type import MetricParamType
from i8_terminal.types.metric_view_param_type import MetricViewParamType
from i8_terminal.types.period_type_param_type import PeriodTypeParamType
from i8_terminal.types.price_period_param_type import PricePeriodParamType
from i8_terminal.types.ticker_param_type import TickerParamType
from i8_terminal.types.user_watchlist_tickers_param_type import (
    UserWatchlistTickersParamType,
)
from i8_terminal.types.user_watchlists_param_type import UserWatchlistsParamType


class I8Completer(ClickCompleter):
    def __init__(self, cli: Callable[[F], Group]) -> None:
        self.cli = cli

    def get_completions(self, document: Document, complete_event: CompleteEvent = None) -> Iterable[Completion]:
        ctx = CommandParser(self.cli).parse(document)
        if not ctx:
            return None

        command = ctx.click_ctx.command

        choices = []
        filter_choices = True

        if ctx.last_option:
            matched_params = [p for p in command.params if isinstance(p, click.Option) and ctx.last_option in p.opts]
            if len(matched_params) > 0:
                matched_param = matched_params[0]
                if type(matched_param.type) is click.types.Choice:
                    for c in matched_param.type.choices:
                        choices.append(Completion(text_type(c), -len(ctx.incomplete)))

                elif type(matched_param.type) in [
                    AutoCompleteChoice,
                    FinancialStatementParamType,
                    PeriodTypeParamType,
                    PricePeriodParamType,
                    FinancialsPeriodParamType,
                    ChartParamType,
                ]:
                    filter_choices = False
                    parts = ctx.incomplete.split(",")
                    incomplete = parts[-1] if len(parts) > 0 else " "
                    for (ticker, name) in matched_param.type.get_suggestions(incomplete if incomplete else " ", True):  # type: ignore
                        choices.append(Completion(text_type(ticker), -len(incomplete), display_meta=name))
                elif type(matched_param.type) in [TickerParamType, UserWatchlistTickersParamType]:
                    filter_choices = False
                    if matched_param.name == "ticker":
                        incomplete = ctx.incomplete
                    else:
                        parts = ctx.incomplete.split(",")
                        incomplete = parts[-1] if len(parts) > 0 else " "
                    for (ticker, name) in matched_param.type.get_suggestions(incomplete if incomplete else " ", True):  # type: ignore
                        choices.append(Completion(text_type(ticker.upper()), -len(incomplete), display_meta=name))
                elif type(matched_param.type) in [MetricParamType, IndicatorParamType]:
                    filter_choices = False
                    parts = ctx.incomplete.split(",")
                    incomplete = parts[-1] if len(parts) > 0 else " "
                    for (metric, name) in matched_param.type.get_suggestions(incomplete if incomplete else " ", True):  # type: ignore
                        choices.append(Completion(text_type(metric.lower()), -len(incomplete), display_meta=name))

                elif type(matched_param.type) is FinancialsIdentifierParamType:
                    param_type = "ticker"
                    filter_choices = False
                    parts = ctx.incomplete.split(",")
                    incomplete = parts[-1] if len(parts) > 0 else " "
                    sub_parts = incomplete.split("-")
                    if len(sub_parts) > 2:
                        param_type = "quarter"
                    elif len(sub_parts) > 1:
                        param_type = "year"
                    else:
                        param_type = "ticker"
                    incomplete = sub_parts[-1] if len(sub_parts) > 0 else " "
                    for (idf, name) in matched_param.type.get_suggestions(
                        incomplete if incomplete else " ", False, param_type
                    ):
                        choices.append(Completion(text_type(idf.upper()), -len(incomplete), display_meta=name))
                elif type(matched_param.type) is UserWatchlistsParamType:
                    filter_choices = False
                    parts = ctx.incomplete.split(",")
                    incomplete = parts[-1] if len(parts) > 0 else " "
                    for (watchlist, desc) in matched_param.type.get_suggestions(
                        incomplete if incomplete else " ", True
                    ):
                        choices.append(Completion(text_type(watchlist), -len(incomplete)))
                elif type(matched_param.type) is MetricViewParamType:
                    filter_choices = False
                    parts = ctx.incomplete.split(",")
                    incomplete = parts[-1] if len(parts) > 0 else " "
                    for (metric_view, desc) in matched_param.type.get_suggestions(incomplete if incomplete else " ", True):  # type: ignore
                        choices.append(Completion(text_type(metric_view), -len(incomplete)))
        else:
            for param in command.params:
                if isinstance(param, click.Option):
                    if not any(o in ctx.used_options for o in param.opts):
                        choices.append(
                            Completion(
                                text_type(max(param.opts, key=len)),
                                -len(ctx.incomplete),
                                display_meta=f"({param.opts[-1]}) {param.help}",
                            )
                        )
                elif isinstance(param, click.Argument):
                    if isinstance(param.type, click.Choice):
                        for choice in param.type.choices:
                            choices.append(Completion(text_type(choice), -len(ctx.incomplete)))

            if isinstance(command, click.MultiCommand):
                for name in command.list_commands(ctx.click_ctx):
                    sub_command = command.get_command(ctx.click_ctx, name)
                    choices.append(
                        Completion(
                            text_type(name),
                            -len(ctx.incomplete),
                            display_meta=getattr(sub_command, "short_help"),
                        )
                    )

        for item in choices:
            if not filter_choices or item.text.startswith(ctx.incomplete):
                yield item



import click
import investor8_sdk
from pandas import DataFrame
from rich.console import Console

from i8_terminal.commands.financials import financials
from i8_terminal.common.cli import pass_command
from i8_terminal.common.financials import available_fin_df2tree
from i8_terminal.common.utils import validate_ticker
from i8_terminal.types.ticker_param_type import TickerParamType


def get_available_financials_df(ticker: str) -> DataFrame:
    available_fins = investor8_sdk.FinancialsApi().get_list_available_standardized_financials(ticker=ticker)
    return DataFrame([d.to_dict() for d in available_fins])


@financials.command()
@click.option("--ticker", "-k", type=TickerParamType(), required=True, callback=validate_ticker, help="Company ticker.")
@pass_command
def coverage(ticker: str) -> None:
    """
    Shows available financial statements of the given company.

    Examples:

    `i8 financials coverage --ticker AAPL`

    """
    console = Console()
    with console.status("Fetching data...", spinner="material"):
        available_fin_df = get_available_financials_df(ticker)
        available_fins_tree = available_fin_df2tree(available_fin_df, ticker)
    console.print(available_fins_tree)



import codecs
import os
from typing import Any, Dict


def read(rel_path: str) -> str:
    """
    Read a file.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version() -> Any:
    """
    Read version from a file.
    """
    for line in read("version.txt").splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def find_dicts_diff(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> Dict[str, Any]:
    result = {}
    for k in dict1:
        if k in dict2:
            if type(dict1[k]) is dict:
                res = find_dicts_diff(dict1[k], dict2[k])
                if res:
                    result[k] = res
            if dict1[k] != dict2[k]:
                result[k] = dict1[k]
        else:
            result[k] = dict1[k]
    for k in dict2:
        if k not in dict1:
            result[k] = dict2[k]
    return result



import re
from datetime import date
from typing import Any, Optional, Union

import arrow
import numpy as np


def make_svg_responsive(svg_str: str) -> str:
    svg_str = re.sub(r'width=".+?"', "", svg_str, 1)
    svg_str = re.sub(r'height=".+?"', "", svg_str, 1)
    svg_str = re.sub(r'style=""', r'style="fill: rgba(0, 0, 0, 0);"', svg_str, 1)  # Make svg transparent

    return svg_str


def format_number(
    m: int,
    unit: Optional[str] = None,
    decimal: int = 2,
    humanize: bool = False,
    colorize: bool = False,
    in_millions: bool = False,
    exportize: Optional[bool] = False,
) -> Optional[Union[str, int]]:
    res: Optional[Union[str, int]] = None
    if m is None or np.isnan(m):
        return "-"

    if exportize:
        return round(m, 2)

    if in_millions and unit in ["usd", "shares"]:
        number = abs(m)
        if number >= 1e6:
            if m < 0:
                res = f"({number // 1e6:,.1f})"
            else:
                res = f"{number // 1e6:,.1f}"

    elif humanize or unit in ["shares", "usdpershare"]:
        number = abs(m)
        if number < 1e3:
            res = f"{m:,.{decimal}f}"
        if number >= 1e3 and number < 1e6:
            res = f"{m / 1e3:,.2f} K"
        if number >= 1e6 and number < 1e9:
            res = f"{m / 1e6:,.2f} M"
        if number >= 1e9 and number < 1e12:
            res = f"{m / 1e9:,.2f} B"
        if number >= 1e12:
            res = f"{m / 1e12:,.2f} T"
    else:
        res = f"{m:,.{decimal}f}"

    if unit == "percentage":
        res = f"{res}%" if m <= 0 else f"+{res}%"

    if unit == "usd":
        res = f"${res}"

    if colorize:
        color = "red" if m <= 0 else "green"
        res = f"[{color}]{res}[/{color}]"

    return res


def format_date(date: date, use_elapsed_format: bool = False, use_precise_format: bool = False) -> Any:
    if use_elapsed_format:
        time_span = arrow.utcnow() - arrow.get(date)
        days = time_span.days
        if days >= 2:
            return arrow.get(date, tzinfo="US/Eastern").strftime("%b %d, %Y %I:%M %p ET")
        hours, remainder = divmod(time_span.seconds, 3600)
        minutes, seconds = divmod(remainder, 60)
        if not use_precise_format:
            if days > 0:
                time = f'{days} day{"s" if days > 1 else ""}'
            elif hours > 0:
                time = f'{hours} hour{"s" if hours > 1 else ""}'
            elif minutes > 0:
                time = f'{minutes} minute{"s" if minutes > 1 else ""}'
            else:
                time = f"{seconds} seconds"

            return f"{time} ago"
        else:
            if days > 0:
                time = f"{days} day, {hours} hour"
            elif hours > 0:
                time = f"{hours} hour, {minutes} minute"
            elif minutes > 0:
                time = f"{minutes} minute, {seconds} second"
            else:
                time = f"{seconds} second"

            return time
    else:
        return arrow.get(date).strftime("%b %d, %Y")


def format_fyq(fyq: str) -> str:
    return f"{fyq[-2:]} {fyq[2:-2]}"


_formatters_map = {
    ("fyq", "console"): lambda x: format_fyq(x),
    ("fyq", "store"): lambda x: format_fyq(x),
    ("number", "console"): lambda x: format_number(x),
    ("number", "store"): lambda x: format_number(x, exportize=True),
    ("price", "console"): lambda x: format_number(x, unit="usd"),
    ("price", "store"): lambda x: round(x, 2),
    ("financial", "console"): lambda x: format_number(x, unit="usd", humanize=True),
    ("financial", "store"): lambda x: format_number(x, exportize=True),
    ("number_int", "console"): lambda x: format_number(x, decimal=0),
    ("number_int", "store"): lambda x: int(x),
    ("perc", "console"): lambda x: format_number(x, decimal=2, unit="percentage", colorize=True),
    ("perc", "store"): lambda x: format_number(x, exportize=True),
    ("date", "console"): lambda x: format_date(x),
    ("date", "store"): lambda x: format_date(x),
    ("str", "store"): lambda x: x,
    ("str", "console"): lambda x: x,
}


def get_formatter(name: str, target: str) -> Any:
    return _formatters_map[(name, target)]



from typing import Optional, cast

import click
from click.types import DateTime
from pandas import DataFrame
from rich.console import Console

from i8_terminal.commands.price import price
from i8_terminal.common.cli import pass_command
from i8_terminal.common.formatting import get_formatter
from i8_terminal.common.layout import df2Table, format_df
from i8_terminal.common.price import get_historical_price_list_df
from i8_terminal.common.utils import export_data, get_period_code, validate_ticker
from i8_terminal.config import APP_SETTINGS
from i8_terminal.types.price_period_param_type import PricePeriodParamType
from i8_terminal.types.ticker_param_type import TickerParamType


def format_hist_price_df(df: DataFrame, target: str) -> DataFrame:
    formatters = {
        "Date": get_formatter("date", target),
        "open": get_formatter("price", target),
        "close": get_formatter("price", target),
        "low": get_formatter("price", target),
        "high": get_formatter("price", target),
        "volume": get_formatter("number_int", target),
        "change_perc": get_formatter("perc", target),
    }
    col_names = {
        "Date": "Date",
        "open": "Open",
        "close": "Close",
        "low": "Low",
        "high": "High",
        "volume": "Volume",
        "change_perc": "Change (%)",
    }
    return format_df(df, col_names, formatters)


@price.command()
@click.option("--ticker", "-k", type=TickerParamType(), required=True, callback=validate_ticker, help="Company ticker.")
@click.option(
    "--period",
    "-p",
    type=PricePeriodParamType(),
    default="1M",
    help="Historical price period.",
)
@click.option("--from_date", "-f", type=DateTime(), help="Histotical price from date.")
@click.option("--to_date", "-t", type=DateTime(), help="Histotical price to date.")
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@pass_command
def list(
    ticker: str, period: str, from_date: Optional[DateTime], to_date: Optional[DateTime], export_path: Optional[str]
) -> None:
    """
    Lists historical prices for a given TICKER.

    Examples:

    `i8 price list --period 3M --ticker AAPL`

    """
    period_code = get_period_code(period.replace(" ", "").upper())
    console = Console()
    with console.status("Fetching data...", spinner="material"):
        df = get_historical_price_list_df([ticker], period_code, cast(str, from_date), cast(str, to_date))

    if export_path:
        df_formatted = format_hist_price_df(df, "store")
        export_data(
            df_formatted,
            export_path,
            column_width=14,
            column_format=APP_SETTINGS["styles"]["xlsx"]["price"]["column"],
        )
    else:
        df_formatted = format_hist_price_df(df, "console")
        table = df2Table(df_formatted)
        console.print(table)



import logging
import os
import shutil
import sys
import uuid
from typing import Any, Dict

import yaml
from mergedeep import merge
from rich.style import Style

from i8_terminal.i8_exception import I8Exception
from i8_terminal.utils import find_dicts_diff

PACKAGE_PATH = os.path.dirname(__file__)
EXECUTABLE_APP_DIR = os.path.join(os.path.dirname(sys.executable))
OS_HOME_PATH = os.path.expanduser("~")
SETTINGS_FOLDER = os.path.join(OS_HOME_PATH, ".i8_terminal")
USER_SETTINGS_PATH = os.path.join(SETTINGS_FOLDER, "user.yml")
APP_SETTINGS_PATH = os.path.join(SETTINGS_FOLDER, "config.yml")
ASSETS_PATH = os.path.join(PACKAGE_PATH, "assets")
I8_TERMINAL_LOGO_URL = "https://www.investoreight.com/media/i8t-chart-logo.png"


def init_settings() -> None:
    if not os.path.exists(SETTINGS_FOLDER):
        try:
            os.mkdir(SETTINGS_FOLDER)
        except:
            logging.error(
                f"Cannot initialize app. Application needs write access to create app directory in the following path: '{OS_HOME_PATH}'"
            )

    if not os.path.exists(USER_SETTINGS_PATH):
        try:
            user_setting = {"app_instance_id": uuid.uuid4().hex}
            with open(USER_SETTINGS_PATH, "w") as f:
                yaml.dump(user_setting, f)
        except:
            logging.error(
                f"Cannot initalize user settings. Make sure you have write access to the path: '{USER_SETTINGS_PATH}'"
            )

    if not os.path.exists(APP_SETTINGS_PATH):
        try:
            app_settings_src_path = os.path.join(PACKAGE_PATH, "config.yml")
            if os.path.exists(app_settings_src_path):
                shutil.copyfile(app_settings_src_path, APP_SETTINGS_PATH)
            else:
                shutil.copyfile(f"{EXECUTABLE_APP_DIR}/config.yml", APP_SETTINGS_PATH)
        except Exception as e:
            logging.error(
                f"Cannot initalize app settings. Make sure you have write access to the path: '{APP_SETTINGS_PATH}'\n {e}"
            )


def load_user_settings() -> Any:
    if not os.path.exists(USER_SETTINGS_PATH):
        return {}
    with open(USER_SETTINGS_PATH, "r") as f:
        return yaml.safe_load(f) or {}


def load_app_settings() -> Any:
    if not os.path.exists(APP_SETTINGS_PATH):
        return {}
    with open(APP_SETTINGS_PATH, "r") as f:
        return yaml.safe_load(f) or {}


def load_latest_app_settings() -> Any:
    app_settings_src_path = os.path.join(PACKAGE_PATH, "config.yml")
    with open(app_settings_src_path, "r") as f:
        return yaml.safe_load(f) or {}


def save_user_settings(data: Dict[str, Any]) -> None:
    current_user_settings = load_user_settings()
    new_user_settings = {**current_user_settings, **data}
    with open(USER_SETTINGS_PATH, "w") as f:
        yaml.dump(new_user_settings, f)


def delete_user_settings() -> None:
    file = open(USER_SETTINGS_PATH, "w")
    file.close()


def restore_user_settings() -> None:
    current_user_settings = load_user_settings()
    restored_user_setting = {"app_instance_id": current_user_settings.get("app_instance_id")}
    with open(USER_SETTINGS_PATH, "w") as f:
        yaml.dump(restored_user_setting, f)


def update_settings() -> None:
    current_app_settings = load_app_settings()
    latest_app_settings = load_latest_app_settings()
    app_new_settings = find_dicts_diff(latest_app_settings, current_app_settings)
    user_new_settings = find_dicts_diff(current_app_settings, latest_app_settings)
    if app_new_settings or user_new_settings:
        new_settings = merge(current_app_settings, latest_app_settings)
        with open(APP_SETTINGS_PATH, "w") as f:
            yaml.dump(new_settings, f)


def get_table_style(profile_name: str = "default") -> Dict[str, Any]:
    styles = APP_SETTINGS["styles"]["table"][profile_name]
    try:
        return {
            "header_style": Style(**styles["header"]),
            "row_styles": [Style(**styles["row"]), Style(**styles["alternate_row"])],
            "show_lines": styles.get("show_lines", False),
        }
    except:
        raise I8Exception(
            "Cannot parse table style settings from the configuration file! Check to see if the configuration file is formatted correctly!"
        )


def is_user_logged_in() -> bool:
    if not USER_SETTINGS.get("i8_core_api_key") or not USER_SETTINGS.get("i8_core_token"):
        return False
    return True


if "USER_SETTINGS" not in globals():
    init_settings()
    update_settings()
    USER_SETTINGS = load_user_settings()
    APP_SETTINGS = load_app_settings()

if not USER_SETTINGS.get("app_instance_id"):
    user_setting = {"app_instance_id": uuid.uuid4().hex}
    save_user_settings(user_setting)
